
/* Chrysalide - Outil d'analyse de fichiers binaires
 * targetableop.c - prototypes pour l'équivalent Python du fichier "arch/targetableop.c"
 *
 * Copyright (C) 2018 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "targetableop.h"


#include <assert.h>
#include <pygobject.h>


#include <i18n.h>


#include <arch/targetableop.h>


#include "processor.h"
#include "vmpa.h"
#include "../access.h"
#include "../helpers.h"
#include "../format/format.h"



/* Obtient l'adresse de la cible visée par un opérande. */
static PyObject *py_targetable_operand_get_addr(PyObject *, PyObject *);



/******************************************************************************
*                                                                             *
*  Paramètres  : self = contenu binaire à manipuler.                          *
*                args = non utilisé ici.                                      *
*                                                                             *
*  Description : Obtient l'adresse de la cible visée par un opérande.         *
*                                                                             *
*  Retour      : Localisation de la cible ou None.                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_targetable_operand_get_addr(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Instance à retourner        */
    PyObject *src_obj;                      /* Objet pour une position     */
    PyObject *format_obj;                   /* Objet pour un format        */
    PyObject *proc_obj;                     /* Objet pour une architecture */
    int ret;                                /* Bilan de lecture des args.  */
    vmpa2t *src;                            /* Version native d'adresse    */
    GBinFormat *format;                     /* Instance GLib du format     */
    GArchProcessor *proc;                   /* Instance GLib de l'archi.   */
    GTargetableOperand *operand;            /* Instance à manipuler        */
    vmpa2t addr;                            /* Localisation à cibler       */
    bool defined;                           /* Cible définie ?             */

    ret = PyArg_ParseTuple(args, "O!O!O!",
                           get_python_vmpa_type(), &src_obj,
                           get_python_binary_format_type(), &format_obj,
                           get_python_arch_processor_type(), &proc_obj);
    if (!ret) return NULL;

    src = get_internal_vmpa(src_obj);
    assert(src != NULL);

    format = G_BIN_FORMAT(pygobject_get(format_obj));
    proc = G_ARCH_PROCESSOR(pygobject_get(proc_obj));

    operand = G_TARGETABLE_OPERAND(pygobject_get(self));
    assert(operand != NULL);

    defined = g_targetable_operand_get_addr(operand, src, format, proc, &addr);

    if (defined)
        result = build_from_internal_vmpa(&addr);
    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_targetable_operand_type(void)
{
    static PyMethodDef py_targetable_operand_methods[] = {
        {
            "get_addr", py_targetable_operand_get_addr,
            METH_VARARGS,
            "get_addr($self, src, format, proc, /)\n--\n\nGet a target address from an operand."
        },
        { NULL }
    };

    static PyGetSetDef py_targetable_operand_getseters[] = {
        { NULL }
    };

    static PyTypeObject py_targetable_operand_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.arch.TargetableOperand",
        .tp_basicsize   = sizeof(PyObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,

        .tp_doc         = "PyChrysalide targetable operand",

        .tp_methods     = py_targetable_operand_methods,
        .tp_getset      = py_targetable_operand_getseters

    };

    return &py_targetable_operand_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.....TargetableOperand'.*
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool ensure_python_targetable_operand_is_registered(void)
{
    PyTypeObject *type;                     /* Type 'TargetableOperand'    */
    PyObject *module;                       /* Module à recompléter        */
    PyObject *dict;                         /* Dictionnaire du module      */

    type = get_python_targetable_operand_type();

    if (!PyType_HasFeature(type, Py_TPFLAGS_READY))
    {
        module = get_access_to_python_module("pychrysalide.arch");

        dict = PyModule_GetDict(module);

        if (!register_interface_for_pygobject(dict, G_TYPE_TARGETABLE_OPERAND, type))
            return false;

    }

    return true;

}
